//
//  SPK_StringToImageStructurePlugIn.m
//  SPK-StringToImageStructure
//
//  Created by Toby Harris on 12/02/2008.
//  Copyright (c) 2008 *spark audio-visual.
//
//  'SPK-StringToImageStructurePlugIn' is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  'SPK-StringToImageStructurePlugIn' is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.


/* It's highly recommended to use CGL macros instead of changing the current context for plug-ins that perform OpenGL rendering */
#import <OpenGL/CGLMacro.h>

#import "SPK_StringToImageStructurePlugIn.h"

#define	kQCPlugIn_Name				@"SPK-StringToImageStructure"
#define	kQCPlugIn_Description		@"Renders a string into per-character images with display metadata. Allows efficient text animation using iterators, as text rendering is cached outside the iterator loop. If no string is supplied on the input port, it will use an alphabet / ascii string. A plug-in by *spark - http://sparkav.co.uk"

@implementation SPK_StringToImageStructurePlugIn

/*
Here you need to declare the input / output properties as dynamic as Quartz Composer will handle their implementation
@dynamic inputFoo, outputBar;
*/

@dynamic inputString, inputFontName, inputFontSize, outputByCharacter, outputByIndex;
@synthesize defaultString;

+ (NSDictionary*) attributes
{
	/*
	Return a dictionary of attributes describing the plug-in (QCPlugInAttributeNameKey, QCPlugInAttributeDescriptionKey...).
	*/
	
	return [NSDictionary dictionaryWithObjectsAndKeys:kQCPlugIn_Name, QCPlugInAttributeNameKey, kQCPlugIn_Description, QCPlugInAttributeDescriptionKey, nil];
}

+ (NSArray*) sortedPropertyPortKeys
{
	/*
	Override this method to specify an optional ordering for property based ports in the user interface.
	*/
	return [NSArray arrayWithObjects:@"inputString",@"inputFontName",@"inputFontSize",nil];
}
			
+ (NSDictionary*) attributesForPropertyPortWithKey:(NSString*)key
{
	/*
	Specify the optional attributes for property based ports (QCPortAttributeNameKey, QCPortAttributeDefaultValueKey...).
	*/
	if([key isEqualToString:@"inputString"])
		return [NSDictionary dictionaryWithObjectsAndKeys:
				@"String", QCPortAttributeNameKey,
				@"", QCPortAttributeDefaultValueKey,
				nil];
	if([key isEqualToString:@"inputFontName"])
		return [NSDictionary dictionaryWithObjectsAndKeys:
				@"Font Name", QCPortAttributeNameKey,
				@"Helvetica", QCPortAttributeDefaultValueKey,
				nil];
	if([key isEqualToString:@"inputFontSize"])
		return [NSDictionary dictionaryWithObjectsAndKeys:
				@"Font Size", QCPortAttributeNameKey,
				[NSNumber numberWithDouble:12], QCPortAttributeDefaultValueKey,
				[NSNumber numberWithDouble:0], QCPortAttributeMinimumValueKey,
				nil];
	if([key isEqualToString:@"outputByCharacter"])
		return [NSDictionary dictionaryWithObject:@"By character" forKey:QCPortAttributeNameKey];
	if([key isEqualToString:@"outputByIndex"])
		return [NSDictionary dictionaryWithObject:@"By index" forKey:QCPortAttributeNameKey];
	return nil;
}

+ (QCPlugInExecutionMode) executionMode
{
	/*
	Return the execution mode of the plug-in: kQCPlugInExecutionModeProvider, kQCPlugInExecutionModeProcessor, or kQCPlugInExecutionModeConsumer.
	*/
	
	return kQCPlugInExecutionModeProcessor;
}

+ (QCPlugInTimeMode) timeMode
{
	/*
	Return the time dependency mode of the plug-in: kQCPlugInTimeModeNone, kQCPlugInTimeModeIdle or kQCPlugInTimeModeTimeBase.
	*/
	
	return kQCPlugInTimeModeNone;
}

- (id) init
{
	if(self = [super init]) {
		/*
		Allocate any permanent resource required by the plug-in.
		*/
		
		// SPK - Load in default string from file in bundle.
		NSString *path = [[NSBundle bundleForClass:[self class]] pathForResource:@"SPK-defaultstring" ofType:@"txt"];
		self.defaultString = [NSString stringWithContentsOfFile:path encoding:NSUTF8StringEncoding error:NULL];
	}
	
	return self;
}

- (void) finalize
{
	/*
	Release any non garbage collected resources created in -init.
	*/
	[super finalize];
}

- (void) dealloc
{
	/*
	Release any resources created in -init.
	*/
	self.defaultString = nil;
	[super dealloc];
}

+ (NSArray*) plugInKeys
{
	/*
	Return a list of the KVC keys corresponding to the internal settings of the plug-in.
	*/
	return [NSArray arrayWithObject:@"defaultString"];
}

- (id) serializedValueForKey:(NSString*)key;
{
	/*
	Provide custom serialization for the plug-in internal settings that are not values complying to the <NSCoding> protocol.
	The return object must be nil or a PList compatible i.e. NSString, NSNumber, NSDate, NSData, NSArray or NSDictionary.
	*/
	
	return [super serializedValueForKey:key];
}

- (void) setSerializedValue:(id)serializedValue forKey:(NSString*)key
{
	/*
	Provide deserialization for the plug-in internal settings that were custom serialized in -serializedValueForKey.
	Deserialize the value, then call [self setValue:value forKey:key] to set the corresponding internal setting of the plug-in instance to that deserialized value.
	*/
	
	[super setSerializedValue:serializedValue forKey:key];
}

- (QCPlugInViewController*) createViewController
{
	/*
	Return a new QCPlugInViewController to edit the internal settings of this plug-in instance.
	You can return a subclass of QCPlugInViewController if necessary.
	*/
	
	return [[QCPlugInViewController alloc] initWithPlugIn:self viewNibName:@"SPK_StringToImageStructureSettings"];
}

// SPK self methods

- (NSFont*)determineFont
{
	// SPK - we know size is > 0, but font is just a text string and may be gibberish
	NSFont* workingFont;
	workingFont = [NSFont fontWithName:self.inputFontName size:self.inputFontSize];
	if (!workingFont) {
		NSLog(@"Font not known, defaulting to system font");
		return [NSFont systemFontOfSize:self.inputFontSize];
	}
	return workingFont;
}

- (NSDictionary*)createCharacterDict:(NSString *)charString 
{
	NSDictionary *attribDict = [NSDictionary dictionaryWithObjectsAndKeys:
								[self determineFont],NSFontAttributeName,
								[NSColor whiteColor],NSForegroundColorAttributeName,
								nil];
	NSSize charSize = [charString sizeWithAttributes:attribDict];
	NSImage *charImage = [[NSImage alloc] initWithSize:charSize];
	[charImage lockFocus];
	[charString drawAtPoint:NSMakePoint(0,0) withAttributes:attribDict];
	[charImage unlockFocus];
	
	return [NSDictionary dictionaryWithObjectsAndKeys:
			charImage,@"Image",
			[NSNumber numberWithDouble:charSize.width],@"Width",
			[NSNumber numberWithDouble:charSize.height],@"Height",
			nil];
}

@end

@implementation SPK_StringToImageStructurePlugIn (Execution)

- (BOOL) startExecution:(id<QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when rendering of the composition starts: perform any required setup for the plug-in.
	Return NO in case of fatal failure (this will prevent rendering of the composition to start).
	*/
	
	return YES;
}

- (void) enableExecution:(id<QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when the plug-in instance starts being used by Quartz Composer.
	*/
}

- (BOOL) execute:(id<QCPlugInContext>)context atTime:(NSTimeInterval)time withArguments:(NSDictionary*)arguments
{
	/*
	Called by Quartz Composer whenever the plug-in instance needs to execute.
	Only read from the plug-in inputs and produce a result (by writing to the plug-in outputs or rendering to the destination OpenGL context) within that method and nowhere else.
	Return NO in case of failure during the execution (this will prevent rendering of the current frame to complete).
	
	The OpenGL context for rendering can be accessed and defined for CGL macros using:
	CGLContextObj cgl_ctx = [context CGLContextObj];
	*/
	
	// SPK - do we need to check for input change to ensure its not rerendering whenever output is asked for?
	NSString *workingString;
	NSUInteger workingLength, i; 
	
	if ([self.inputString length] > 0) {
		workingString = self.inputString;
	} else {
		workingString = self.defaultString;
	}
	workingLength = [workingString length];
	
	NSMutableArray *byIndexArray = [NSMutableArray arrayWithCapacity:workingLength];
	NSMutableDictionary *byCharDictionary = [NSMutableDictionary dictionaryWithCapacity:1];
	double centeredXPos;
	double runningXPos = 0;
	for (i = 0; i < workingLength; i++) {
		// get character
		NSString *character = [workingString substringWithRange:NSMakeRange(i,1)];
		
		// generate 'by character'
		NSDictionary *characterDict = [byCharDictionary objectForKey:character];
		if (!characterDict) {
		characterDict = [self createCharacterDict:character];
		[byCharDictionary setObject:characterDict forKey:character];
		}
		
		// generate 'by index' using characterDict
		centeredXPos = runningXPos + [[characterDict objectForKey:@"Width"] doubleValue] / 2;
		runningXPos += [[characterDict objectForKey:@"Width"] doubleValue];
		NSDictionary *characterInString = [NSDictionary dictionaryWithObjectsAndKeys:
										   character,@"Character",
										   [characterDict objectForKey:@"Image"],@"Image",
										   [characterDict objectForKey:@"Width"],@"Width",
										   [characterDict objectForKey:@"Height"],@"Height",
										   [NSNumber numberWithDouble:centeredXPos],@"xPos",
										   nil];
		[byIndexArray addObject:characterInString];
	}
	self.outputByCharacter = byCharDictionary;
	self.outputByIndex = byIndexArray;
	return YES;
}

- (void) disableExecution:(id<QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when the plug-in instance stops being used by Quartz Composer.
	*/
}

- (void) stopExecution:(id<QCPlugInContext>)context
{
	/*
	Called by Quartz Composer when rendering of the composition stops: perform any required cleanup for the plug-in.
	*/
}

@end
